@@ -1,4 +1,6 @@
 
+use XBase::Memo;
+
 =head1 NAME
 
 XBase - Perl module for reading and writing the dbf files
@@ -18,7 +20,7 @@ use XBase::Base;		# will give us general methods
 use vars qw( $VERSION $errstr $CLEARNULLS @ISA );
 
 @ISA = qw( XBase::Base );
-$VERSION = '0.0694';
+$VERSION = '0.155';
 $CLEARNULLS = 1;		# Cut off white spaces from ends of char fields
 
 *errstr = \$XBase::Base::errstr;
@@ -34,16 +36,23 @@ sub open
 	my %options;
 	if (scalar(@_) % 2) { $options{'name'} = shift; }	
 	$self->{'openoptions'} = { %options, @_ };
-	my $filename = $self->{'openoptions'}{'name'};
+
+	my %locoptions;
+	@locoptions{ qw( name readonly ignorememo ) } = @{$self->{'openoptions'}}{ qw( name readonly ignorememo ) };
+	my $filename = $locoptions{'name'};
+	if ($filename eq '-')
+		{ return $self->SUPER::open(%locoptions); }
 	for my $ext ('', '.dbf', '.DBF')
 		{
 		if (-f $filename.$ext)
 			{
+			$locoptions{'name'} = $filename.$ext;
 			$self->NullError();
-			return $self->SUPER::open($filename.$ext);
+			return $self->SUPER::open(%locoptions);
 			}
 		}
-	return $self->SUPER::open($filename);	# for nice error message
+	$locoptions{'name'} = $filename;	
+	return $self->SUPER::open(%locoptions);	# for nice error message
 	}
 # We have to provide way to fill up the object upon open
 sub read_header
@@ -52,7 +61,7 @@ sub read_header
 	my $fh = $self->{'fh'};
 
 	my $header;				# read the header
-	$fh->read($header, 32) == 32 or do
+	$self->read($header, 32) == 32 or do
 		{ __PACKAGE__->Error("Error reading header of $self->{'filename'}: $!\n"); return; };
 
 	@{$self}{ qw( version last_update num_rec header_len record_len ) }
@@ -62,15 +71,14 @@ sub read_header
 	my ($names, $types, $lengths, $decimals) = ( [], [], [], [] );
 	my ($unpacks, $readproc, $writeproc) = ( [], [], [] );
 	my $lastoffset = 1;
-	
-	while (tell($fh) < $header_len - 1)	# read the field desc's
+
+	while ($self->tell() < $header_len - 1)	# read the field desc's
 		{
 		my $field_def;
-		my $read = $fh->read($field_def, 32);
-		last if substr($field_def, 0, 1) eq "\r";
-						# we have found the terminator
-		
-		if ($read != 32)	
+		$self->read($field_def, 1);
+		last if $field_def eq "\r";	# we have found the terminator
+		my $read = $self->read($field_def, 31, 1);
+		if ($read != 31)	
 			{ __PACKAGE__->Error("Error reading field description: $!\n"); return; };
 
 		my ($name, $type, $length, $decimal)
@@ -82,9 +90,9 @@ sub read_header
 			# fixup for char length > 256
 			if ($decimal and not $self->{'openoptions'}{'nolongchars'})
 				{ $length += 256 * $decimal; $decimal = 0; }
-			$rproc = sub { my $value = shift;
+			$rproc = sub { my $value = $_[0];
 				if ($self->{'ChopBlanks'})
-					{ $value =~ s/\s+$//; } ### $value =~ s/^\s+//; }
+					{ $value =~ s/\s+$//; }
 				return $value;
 				( $value eq '' ? undef : $value ); };
 			$wproc = sub { my $value = shift;
@@ -114,22 +122,44 @@ sub read_header
 			$rproc = sub { unpack 'V', shift; };
 			$wproc = sub { pack 'V', shift; };
 			}
-		elsif ($type =~ /^[MGBP]$/)	# memo fields
+		elsif ($type eq 'B')		# Fox double
+			{
+			$rproc = sub { unpack 'd', reverse scalar shift; };
+			$wproc = sub { reverse scalar pack 'd', shift; };
+			}
+		elsif ($type =~ /^[MGP]$/)	# memo fields
 			{
 			my $memo = $self->{'memo'};
 			if (not defined $memo and not $self->{'openoptions'}{'ignorememo'})
 				{ $memo = $self->{'memo'} = $self->init_memo_field() or return; }
-			if (defined $memo)
+			if (defined $memo and $length == 10)
+				{
+				if (ref $memo eq 'XBase::Memo::Apollo') {
+					$rproc = sub { $memo->read_record(shift); };
+					$wproc = sub { $memo->write_record(shift); };
+					}
+				else {
+					$rproc = sub {
+						my $value = shift;
+						return if not $value =~ /\d/ or $value < 0;
+						$memo->read_record($value - 1) if defined $memo;
+						};
+					$wproc = sub {
+						my $value = $memo->write_record(-1, $type, $_[0]) if defined $memo and defined $_[0];
+						sprintf '%*.*s', $length, $length,
+							(defined $value ? $value + 1: ''); };
+					}
+				}
+			elsif (defined $memo and $length == 4)
 				{
 				$rproc = sub {
-					my $value = shift;
-					return undef unless $value =~ /\d/;
-					$memo->read_record($value - 1) if defined $memo;
+					my $val = unpack('V', $_[0]) - 1;
+					return if $val < 0;
+					$memo->read_record($val) if defined $memo;
 					};
 				$wproc = sub {
 					my $value = $memo->write_record(-1, $type, shift) if defined $memo;
-					sprintf '%*.*s', $length, $length,
-						(defined $value ? $value + 1: ''); };
+					pack 'V', (defined $value ? $value + 1: 0); };
 				}
 			else
 				{
@@ -137,6 +167,36 @@ sub read_header
 				$wproc = sub { ' ' x $length; };
 				}
 			}
+		elsif ($type eq 'T')	# time fields
+			{
+			$rproc = sub {
+				my ($day, $time) = unpack 'VV', $_[0];
+
+
+				my $localday = $day - 2440588;
+				my $localtime = $localday * 24 * 3600;
+				$localtime += $time / 1000;
+### print STDERR "day,time: ($day,$time -> $localtime)\n";
+				return $localtime;
+
+				my $localdata = "[$localday] $localtime: @{[localtime($localtime)]}";
+
+				my $usec = $time % 1000;
+				my $hour = int($time / 3600000);
+				my $min = int(($time % 3600000) / 60000);
+				my $sec = int(($time % 60000) / 1000);
+				return "$day($localdata)-$hour:$min:$sec.$usec";
+				};
+			$wproc = sub {
+				my $localtime = shift;
+				my $day = int($localtime / (24 * 3600)) + 2440588;
+				my $time = int(($localtime % (3600 * 24)) * 1000);
+
+### print STDERR "day,time: ($localtime -> $day,$time)\n";
+
+				return pack 'VV', $day, $time;	
+				}
+			}
 		$name =~ s/[\000 ].*$//s;
 		$name = uc $name;		# no locale yet
 		push @$names, $name;
@@ -174,14 +234,14 @@ sub init_memo_field
 	if (defined $self->{'openoptions'}{'memofile'})
 		{ return XBase::Memo->new($self->{'openoptions'}{'memofile'}, %options); }
 	
-	my $memo;
-	my $memoname = $self->{'filename'};
-	$memoname =~ s/\.DBF$/.FPT/ or $memoname =~ s/(\.dbf)?$/.fpt/;
-	$memo = XBase::Memo->new($memoname, %options) and return $memo;
-	
-	$memoname = $self->{'filename'};
-	$memoname =~ s/\.DBF$/.DBT/ or $memoname =~ s/(\.dbf)?$/.dbt/;
-	$memo = XBase::Memo->new($memoname, %options) and return $memo;
+	for (qw( dbt DBT fpt FPT smt SMT dbt ))
+		{
+		my $memo;
+		my $memoname = $self->{'filename'};
+		($memoname =~ s/\.dbf$/.$_/i or $memoname =~ s/(\.dbf)?$/.$_/i)
+			and $memo = XBase::Memo->new($memoname, %options)
+			and return $memo;
+		}
 	return;
 	}
 # Close the file (and memo)
@@ -237,7 +297,7 @@ sub get_header_info
 	{
 	my $self = shift;
 	my $hexversion = sprintf '0x%02x', $self->version;
-	my $longversion = $self->get_version_info;
+	my $longversion = $self->get_version_info()->{'string'};
 	my $printdate = $self->get_last_change;
 	my $numfields = $self->last_field() + 1;
 	my $result = sprintf <<"EOF";
@@ -277,20 +337,24 @@ sub get_version_info
 	{
 	my $version = shift;
 	$version = $version->version() if ref $version;
-	my ($vbits, $dbtflag, $memo, $sqltable) = (0, 0, 0, 0);
-	if ($version == 3)	{ $vbits = 3; }
-	elsif ($version == 0x83)	{ $vbits = 3; $memo = 0; $dbtflag = 1;}
-	else {
-		$vbits = $version & 0x07;
-		$dbtflag = ($version >> 7) & 1;
-		$memo = ($version >> 3) & 1;
-		$sqltable = ($version >> 4) & 0x07;
-		}
-	
-	my $result = "ver. $vbits";
-	if ($memo)	{ $result .= ' with memo file'; }
-	elsif ($dbtflag)	{ $result .= ' with DBT file'; }
-	$result .= ' containing SQL table' if $sqltable;
+	my $result = {};
+	$result->{'vbits'} = $version & 0x07;
+	if ($version == 0x30 or $version == 0xf5)
+		{ $result->{'vbits'} = 5; $result->{'foxpro'} = 1; }
+	elsif ($version & 0x08)
+		{ $result->{'vbits'} = 4; $result->{'memo'} = 1; }
+	elsif ($version & 0x80)
+		{ $result->{'dbt'} = 1; }
+
+	my $string = "ver. $result->{'vbits'}";
+	if (exists $result->{'foxpro'})
+		{ $string .= " (FoxPro)"; }
+	if (exists $result->{'memo'})
+		{ $string .= " with memo file"; }
+	elsif (exists $result->{'dbt'})
+		{ $string .= " with DBT file"; }
+	$result->{'string'} = $string;
+
 	$result;
 	}
 
@@ -308,15 +372,66 @@ sub dump_records
 		$outkey =~ s/[^a-z]//g;
 		$options{$outkey} = $value;
 		}
+	my ($rs, $fs, $undef, $fields, $table)
+				= @options{ qw( rs fs undef fields table ) };
+	if (defined $table)
+		{
+		eval 'use Data::ShowTable';
+		if ($@) {
+			warn "You requested table output format but the module Data::ShowTable doesn't\nseem to be installed correctly. Falling back to standard\n";
+			$table = undef;
+			}
+		else
+			{ delete $options{'rs'}; delete $options{'fs'}; }
+		}
 
-	my ($rs, $fs, $undef, $fields) = @options{ qw( rs fs undef fields ) };
 	my @fields = ();
-	@fields = @$fields if defined $fields;
+	if (defined $fields)
+		{
+		if (ref $fields eq 'ARRAY') { @fields = @$fields; }
+		else
+			{
+			@fields = split /\s*,\s*/, $fields;
+			my $i = 0;
+			while ($i < @fields)
+				{
+				if (defined $self->field_name_to_num($fields[$i]))
+					{ $i++; }
+				elsif ($fields[$i] =~ /^(.*)-(.*)/)
+					{
+					my @allfields = $self->field_names;
+					my ($start, $end) = ($1, $2);
+					if ($start eq '')
+						{ $start = $allfields[0]; }
+					if ($end eq '')
+						{ $end = $allfields[$#allfields]; }
+					$start = $self->field_name_to_num($start);
+					$end = $self->field_name_to_num($end);
+					unless (defined $start and defined $end)
+						{ $start = 0; $end = -1; }
+					
+					splice @fields, $i, 1, @allfields[$start .. $end];
+					}
+				}
+			}
+		}
 
 	my $cursor = $self->prepare_select(@fields);
 	my @record;
-	while (@record = $cursor->fetch())
-		{ print join($fs, map { defined $_ ? $_ : $undef } @record), $rs; }
+	if (defined $table)
+		{
+		local $^W = 0;
+		&ShowBoxTable( $cursor->names(), [], [],
+			sub {
+				if ($_[0]) { $cursor->rewind(); }
+				else { $cursor->fetch() }
+				});
+		}
+	else
+		{
+		while (@record = $cursor->fetch)
+			{ print join($fs, map { defined $_ ? $_ : $undef } @record), $rs; }
+		}
 	1;
 	}
 
@@ -515,7 +630,7 @@ sub create
 		if (not defined $length)		# defaults
 			{
 			if ($type eq "C")	{ $length = 64; }
-			elsif ($type eq "D")	{ $length = 8; }
+			elsif ($type =~ /^[TD]$/)	{ $length = 8; }
 			elsif ($type =~ /^[NF]$/)	{ $length = 8; }
 			}
 						# force correct lengths
@@ -587,14 +702,26 @@ sub drop
 sub locksh
 	{
 	my $self = shift;
-	$self->SUPER::locksh;
-	$self->{'memo'}->locksh() if defined $self->{'memo'};
+	my $ret = $self->SUPER::locksh or return;
+	if (defined $self->{'memo'}) {
+		unless ($self->{'memo'}->locksh()) {
+			$self->SUPER::unlock;
+			return;
+			}
+		}
+	$ret;
 	}
 sub lockex
 	{
 	my $self = shift;
-	$self->SUPER::lockex;
-	$self->{'memo'}->lockex() if defined $self->{'memo'};
+	my $ret = $self->SUPER::lockex or return;
+	if (defined $self->{'memo'}) {
+		unless ($self->{'memo'}->lockex()) {
+			$self->SUPER::unlock;
+			return;
+			}
+		}
+	$ret;
 	}
 sub unlock
 	{
@@ -617,14 +744,28 @@ sub prepare_select
 		# object, recno, field numbers, field names
 	}
 
+sub prepare_select_nf
+	{
+	my $self = shift;
+	my @fieldnames = $self->field_names;
+	if (@_) { @fieldnames = @fieldnames[ @_ ] }
+	return $self->prepare_select(@fieldnames);
+	}
+
 sub prepare_select_with_index
 	{
 	my ($self, $file) = ( shift, shift );
+	my @tagopts = ();
+	if (ref $file eq 'ARRAY') {		### this is suboptimal
+				### interface but should suffice for the moment
+		@tagopts = ( 'tag' => $file->[1]);
+		$file = $file->[0];
+		}
 	my $fieldnames = [ @_ ];
 	if (not @_) { $fieldnames = [ $self->field_names ] };
 	my $fieldnums = [ map { $self->field_name_to_num($_); } @$fieldnames ];
 	require XBase::Index;
-	my $index = new XBase::Index $file or
+	my $index = new XBase::Index $file, 'dbf' => $self, @tagopts or
 		do { $self->Error(XBase->errstr); return; };
 	$index->prepare_select or
 		do { $self->Error($index->errstr); return; };
@@ -666,6 +807,16 @@ sub last_fetched
 	{ shift->[1]; }
 sub table
 	{ shift->[0]; }
+sub names
+	{ shift->[3]; }
+sub rewind
+	{ shift->[1] = undef; '0E0'; }
+
+sub attach_index {
+	my $self = shift;
+	require XBase::Index;
+
+	}
 
 package XBase::IndexCursor;
 use vars qw( @ISA );
@@ -707,14 +858,11 @@ __END__
 
 This module can read and write XBase database files, known as dbf in
 dBase and FoxPro world. It also reads memo fields from the dbt and fpt
-files, if needed. Module XBase provides simple native interface to
-XBase files. For DBI compliant database access, see the DBD::XBase
-and DBI modules.
-
-B<New:> There is a support for B<ndx> and B<ntx> index files
-available. Check the B<prepare_select_with_index> method in this man
-page, or eg/use_index if you are brave and want to help me debugging
-the code.
+files, if needed. An alpha code of reading index support for ndx, ntx,
+mdx, idx and cdx is available for testing -- see the DBD::Index(3) man
+page. Module XBase provides simple native interface to XBase files.
+For DBI compliant database access, see the DBD::XBase and DBI modules
+and their man pages.
 
 The following methods are supported by XBase module:
 
@@ -739,9 +887,11 @@ By default it's the name of the dbf file, with extension dbt or fpt.
 B<ignorememo> ignore memo file at all. This is usefull if you've lost
 the dbt file and you do not need it. Default is false.
 
-B<memosep> separator of memo records in the dBase III dbt files, to
-read files created by broken clients, that put there something else
-than the default C<"\x1a\x1a">.
+B<memosep> separator of memo records in the dBase III dbt files. The
+standard says it should be C<"\x1a\x1a">. There are however
+implamentations that only put in one C<"\x1a">. XBase.pm tries to
+guess which is valid for your dbt but if it fails, you can tell it
+yourself.
 
 B<nolongchars> prevents XBase to treat the decimal value of character
 fields as high byte of the length -- there are some broken products
@@ -770,8 +920,8 @@ If you call B<create> using class name (XBase), you have to (besides
 B<name>) also specify another four values, each being a reference
 to list: B<field_names>, B<field_types>, B<field_lengths> and
 B<field_decimals>. The field types are specified by one letter
-strings (C, N, L, D). If you set some value as undefined, create will
-make it into some reasonable default.
+strings (C, N, L, D, ...). If you set some value as undefined, create
+will make it into some reasonable default.
 
     my $newtable = $table->create("name" => "copy.dbf");
 	
@@ -911,8 +1061,10 @@ the following B<fetch> will return all fields.
 =item prepare_select_with_index
 
 The first parameter is the file name of the index file, the rest is
-as above. The B<fetch> will then return records in the ascending
-order, according to the index.
+as above. For index types that can hold more index structures in on
+file, use arrayref instead of the file name and in that array include
+file name and the tag name. The B<fetch> will then return records in
+the ascending order, according to the index.
 
 =back
 
@@ -953,6 +1105,8 @@ Examples of using cursors:
 
     my $table = new XBase "employ.dbf";
     my $cur = $table->prepare_select_with_index("empid.ndx");
+    ## my $cur = $table->prepare_select_with_index(
+		["empid.cdx", "ADDRES"], "id", "address");
     $cur->find_eq(1097);
     while (my $hashref = $cur->fetch_hashref
 			and $hashref->{"ID"} == 1097)
@@ -1029,17 +1183,31 @@ information about the file and about the fields.
 Module XBase::Base(3) defines some basic functions that are inherited
 by both XBase and XBase::Memo(3) module.
 
-=head1 MEMO, INDEX, LOCKS
+=head1 DATA TYPES
+
+The character fields are returned "as is". No charset or other
+translation is done. The numbers are converted to Perl numbers. The
+date fields are returned as 8 character string of the 'YYYYMMDD' form
+and when inserting the date, you again have to provide it in this
+form. No checking for the validity of the date is done. The datetime
+field is returned in the number of seconds since 1970/1/1, possibly
+with decimal part (since it allows precision up to 1/1000 s). To get
+the fields, use the gmtime (or similar) Perl function.
 
 If there is a memo field in the dbf file, the module tries to open
-file with the same name but extension dbt or fpt. It uses module
+file with the same name but extension dbt, fpt or smt. It uses module
 XBase::Memo(3) for this. It reads and writes this memo field
-transparently (you do not know about it).
+transparently (you do not know about it) and returns the data as
+single scalar.
 
-B<New:> There is a small read only support available for ndx and ntx
-index files. Please see the eg/use_index file in the distribution for
-examples and ideas. Send me examples of your data files and
-suggestions for interface if you need indexes.
+=head1 INDEX, LOCKS
+
+B<New:> A support for ndx, ntx, mdx, idx and cdx index formats is
+available with alpha status for testing. Some of the formats are
+already rather stable (ndx). Please read the XBase::Index(3) man page
+and the eg/use_index file in the distribution for examples and ideas.
+Send me examples of your data files and suggestions for interface if
+you need indexes.
 
 General locking methods are B<locksh>, B<lockex> and B<unlock> for
 shared lock, exclusive lock and unlock. They call flock but you can
@@ -1056,17 +1224,33 @@ Thanks a lot.
 
 =head1 VERSION
 
-0.0694
+0.155
 
 =head1 AUTHOR
 
-(c) 1997--1998 Jan Pazdziora, adelton@fi.muni.cz,
+(c) 1997--1999 Jan Pazdziora, adelton@fi.muni.cz,
 http://www.fi.muni.cz/~adelton/ at Faculty of Informatics, Masaryk
 University in Brno, Czech Republic
 
+All rights reserved. This package is free software; you can
+redistribute it and/or modify it under the same terms as Perl itself.
+
+=head1 THANKS
+
+Many people have provided information, test files, test results and
+patches. This project would not be so great without them. See the
+Changes file for (I hope) complete list. Thank you all, guys!
+
+Special thanks go to Erik Bachmann for his great page about the
+file structures; to Frans van Loon, William McKee, Randy Kobes and
+Dan Albertsson for longtime cooperation and many emails we've
+exchanged when fixing and polishing the modules' behaviour; and to
+Dan Albertsson for providing support for the project.
+
 =head1 SEE ALSO
 
-perl(1); DBD::XBase(3) and DBI(3) for DBI interface; dbfdump(1)
+perl(1); XBase::FAQ(3); DBD::XBase(3) and DBI(3) for DBI interface;
+dbfdump(1)
 
 =cut
 
